/*
 * Final project, CMPS 160, Fall 1999
 * Ben Gertzfield <ben@cse.ucsc.edu>
 *
 * This program is released under the GNU GPL, version 2 or later.
 */

/* Main routines */

#include <X11/Xlib.h>
#include <X11/keysym.h>
#include <gtk/gtk.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>

#ifdef HAVE_XMESAFINDBUFFER
#include <GL/xmesa.h>
#endif

#include <pthread.h>

#ifdef HAVE_SCHED_SETSCHEDULER
#include <sched.h>
#endif

#include <sys/time.h>
#include <unistd.h>

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <xmms/plugin.h>
#include <xmms/xmmsctrl.h>
#include <xmms/util.h>
#include <xmms/configfile.h>

#include "prog.h"
#include "interface.h"


static Display *dpy = NULL;
static Colormap colormap = 0;
static GLXContext glxcontext = NULL;
static Window window = 0;
static Atom wm_delete_window_atom;
static pthread_t draw_thread;
static int screen_number = 0, root_x = 640, root_y = 480;
static struct timeval last_time;

GLfloat gkos_winw = INIT_WINW, gkos_winh = INIT_WINH;

static gboolean grabbed_pointer = FALSE;
gboolean gkos_going = FALSE;

static void geekosphere_init(void);
static void geekosphere_cleanup(void);
static void geekosphere_playback_start(void);
static void geekosphere_playback_stop(void);
static void geekosphere_render_freq(gint16 data[2][256]);

pthread_mutex_t draw_mutex = PTHREAD_MUTEX_INITIALIZER;
static void idle(void);

static GLdouble frames_per_second(void);

VisPlugin geekosphere_vp =
{
    NULL,
    NULL,
    0,
    "Geekosphere " VERSION " (OpenGL)",
    0,
    1,
    geekosphere_init,		/* init */
    geekosphere_cleanup,	/* cleanup */
    geekosphere_about,		/* about */
    geekosphere_configure,	/* configure */
    NULL,			/* disable */
    geekosphere_playback_start,	/* playback_start */
    geekosphere_playback_stop,	/* playback_stop */
    NULL,			/* render_pcm */
    geekosphere_render_freq,	/* render_freq */
};

VisPlugin *get_vplugin_info(void)
{
    return &geekosphere_vp;
}

static gint disable_func(gpointer data)
{
    geekosphere_vp.disable_plugin(&geekosphere_vp);
    return FALSE;
}

static GLdouble frames_per_second(void)
{
    struct timeval this_time;
    long diff_secs, diff_usecs;
    GLdouble fps;

    gettimeofday(&this_time, NULL);

    diff_secs = this_time.tv_sec - last_time.tv_sec;

    if (diff_secs == 0)
	diff_usecs = this_time.tv_usec - last_time.tv_usec;
    else {
	if (this_time.tv_usec >= last_time.tv_usec) {
	    diff_usecs = diff_secs * 1000000 + (this_time.tv_usec - 
						last_time.tv_usec);
	} else {
	    diff_usecs = (diff_secs - 1) * 1000000 + 
		((this_time.tv_usec + 1000000) - last_time.tv_usec);
	}
    }
	    
    last_time = this_time;
    
    fps = 1000000.0 / diff_usecs;

    return fps;
}

void *draw_thread_func(void *arg)
{
    Bool configured = FALSE;

#ifdef HAVE_SCHED_SETSCHEDULER
    if(xmms_check_realtime_priority()) {
	struct sched_param sparam;
	sparam.sched_priority = sched_get_priority_max(SCHED_OTHER);
	pthread_setschedparam(pthread_self(), SCHED_OTHER, &sparam);
    }
#endif
	
    while (gkos_going) {
	while (XPending(dpy)) {
	    XEvent event;
	    KeySym keysym;
	    char buf[16];
	    
	    XNextEvent(dpy, &event);
	    switch(event.type) {
	    case ConfigureNotify:
		glViewport(0, 0, (GLsizei) event.xconfigure.width, (GLsizei)  event.xconfigure.height);
		gkos_winw = event.xconfigure.width;
		gkos_winh = event.xconfigure.height;

		if (!gkos_config->tdfx_mode) {
		    gkos_config->winw = gkos_winw;
		    gkos_config->winh = gkos_winh;
		}

		gkos_setup_projection(gkos_cur_camera);
		if (gkos_config->tdfx_mode && !grabbed_pointer) {
		    XGrabPointer(dpy, window, True, ButtonPressMask,
				 GrabModeAsync, GrabModeAsync, window,
				 None, CurrentTime);
		    grabbed_pointer = TRUE;
		}
		configured = TRUE;
		break;
		
	    case KeyPress:
		XLookupString (&event.xkey, buf, 16, &keysym, NULL);
		switch(keysym) {
		case XK_Escape:
		case XK_q:
		case XK_Q:
		    /* Ugly hack to get the disable_plugin call in the main thread. */
		    GDK_THREADS_ENTER();
		    gtk_idle_add(disable_func, NULL);
		    GDK_THREADS_LEAVE();
		    break;
		case XK_Right:
		case XK_KP_Right:
		    xmms_remote_playlist_next(geekosphere_vp.xmms_session);
		    break;
		case XK_Left:
		case XK_KP_Left:
		    xmms_remote_playlist_prev(geekosphere_vp.xmms_session);
		    break;
		case XK_s:
		case XK_S:
		    xmms_remote_stop(geekosphere_vp.xmms_session);
		    break;
		case XK_p:
		case XK_P:
		    xmms_remote_play(geekosphere_vp.xmms_session);
		    break;
		case XK_space:
		case XK_KP_Begin:
		    xmms_remote_pause(geekosphere_vp.xmms_session);
		    break;
		case ClientMessage:
		    if ((Atom)event.xclient.data.l[0] == wm_delete_window_atom) {
			GDK_THREADS_ENTER();
			gtk_idle_add(disable_func, NULL);
			GDK_THREADS_LEAVE();
		    }
		    break;
		}
	    }
	}

	if(configured)
	    idle();
    }
    pthread_exit(NULL);
}

/*
 * Create a GLX window with the specified width and height.
 *
 * GLX code based on example opengl_spectrum XMMS plugin.
 *
 */

static Window create_window(int width, int height)
{	
    int attr_list[] = { GLX_RGBA,
			GLX_DEPTH_SIZE, 16, 
			GLX_DOUBLEBUFFER,
			None };
    XSetWindowAttributes attr;
    unsigned long mask;
    Window root, win;
    XVisualInfo *visinfo;
    Atom wm_protocols[1];
    
    screen_number = DefaultScreen(dpy);
    root = RootWindow(dpy, screen_number);
    root_x = DisplayWidth(dpy, screen_number);
    root_y = DisplayHeight(dpy, screen_number);

    visinfo = glXChooseVisual(dpy, screen_number, attr_list);
    
    if(!visinfo)
	return 0;
    
    attr.background_pixel = 0;
    attr.border_pixel = 0;
    attr.colormap = colormap = XCreateColormap(dpy, root, visinfo->visual, AllocNone);
    attr.event_mask = StructureNotifyMask | KeyPressMask;
    mask = CWBackPixel | CWBorderPixel | CWColormap | CWEventMask;
    win = XCreateWindow(dpy, root, 0, 0, width, height,
			0, visinfo->depth, InputOutput,
			visinfo->visual, mask, &attr);

    XmbSetWMProperties(dpy, win, "Geekosphere " VERSION,
		       "Geekosphere " VERSION, NULL, 0, NULL, NULL,
		       NULL);
    wm_delete_window_atom = wm_protocols[0] = XInternAtom(dpy,
							  "WM_DELETE_WINDOW",
							  False);

    XSetWMProtocols(dpy, win, wm_protocols, 1);

    glxcontext = glXCreateContext(dpy, visinfo, NULL, True);

    glXMakeCurrent(dpy, win, glxcontext);

    return win;
}

static void start_display(void)
{
#ifdef HAVE_XMESAFINDBUFFER
    if (gkos_config->tdfx_mode)
	setenv("MESA_GLX_FX", "fullscreen", TRUE);
    else
	unsetenv("MESA_GLX_FX");
#endif

    dpy = XOpenDisplay(NULL);
    window = create_window(gkos_config->winw, gkos_config->winh);

    XMapWindow(dpy, window);

    if (gkos_config->tdfx_mode)
	XMoveResizeWindow(dpy, window, 0, 0, root_x, root_y);

    glClearColor(0.0, 0.0, 0.0, 0.0);
    glColor3f(1.0, 1.0, 1.0);
    glDepthFunc(GL_LESS);
    glShadeModel(GL_SMOOTH);
    glEnable(GL_LIGHTING);
    glEnable(GL_AUTO_NORMAL);
    glEnable(GL_NORMALIZE);
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

    srand(time(NULL));
    gettimeofday(&last_time, NULL);

    gkos_going = 1;

    gkos_new_camera();
    gkos_init_lights();
    gkos_init_textures();
    gkos_init_objects();

    //gkos_fun_nurb();
    //gkos_cobb_nurbsphere();
    gkos_sweep_sphere(3.0);
    //gkos_sweep_chessman();

    pthread_create(&draw_thread, NULL, draw_thread_func, NULL);
}

static void stop_display(void)
{
    if (gkos_going) {
	gkos_going = FALSE;
	pthread_join(draw_thread, NULL);
    }

    if (window) {
	if (grabbed_pointer) {
	    XUngrabPointer(dpy, CurrentTime);
	    grabbed_pointer = FALSE;
	}

	XDestroyWindow(dpy, window);
	window = 0;
    }

    if (glxcontext) {
	glXMakeCurrent(dpy, 0, NULL);
	glXDestroyContext(dpy, glxcontext);
	glxcontext = NULL;
    }	

    if (colormap) {
	XFreeColormap(dpy, colormap);
	colormap = 0;
    }

    if (dpy) {
	XCloseDisplay(dpy);
	dpy = NULL;
    }

    while (gkos_first_object != NULL)
	gkos_pop_object();

    while (gkos_first_light != NULL)
	gkos_pop_light();

    while (gkos_first_camera != NULL)
	gkos_pop_camera();

}

static void geekosphere_init(void)
{
    gkos_init_configuration();

    if (!gkos_config->tdfx_mode)
	start_display();
}	

static void geekosphere_cleanup(void)
{
    stop_display();
    pthread_mutex_destroy(&draw_mutex);

    if (!gkos_config->tdfx_mode)
	gkos_write_config();
}

static void geekosphere_playback_start(void)
{
    if (gkos_config->tdfx_mode) {
	if (window)
	    stop_display();
	start_display();
    }
}

static void geekosphere_playback_stop(void)
{
    gkos_reset_all_objects();

    if (gkos_config->tdfx_mode)
	stop_display();
}

static void geekosphere_render_freq(gint16 data[2][256])
{
    struct object *obj = gkos_first_object;
    
    while (obj != NULL) {
	if (obj->update == NULL) {
	    obj = obj->next;
	    continue;
	}

	(*obj->update)(obj, data[0]);
	obj = obj->next;
    }
}

void idle()
{
    struct object *obj = gkos_first_object;
    struct light *light = gkos_first_light;
    GLdouble fps, target;

    while (obj != NULL) {
	if (obj->idle == NULL) {
	    obj = obj->next;
	    continue;
	}

	(*obj->idle)(obj);
	obj = obj->next;
    }

    while (light != NULL) {
	if (light->idle == NULL) {
	    light = light->next;
	    continue;
	}

	(*light->idle)(light);
	light = light->next;
    }

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    gkos_enable_all_lights();
    gkos_draw_all_objects();
    
    if (gkos_config->calc_fps || gkos_config->max_fps) {
	fps = frames_per_second();

	if (gkos_config->max_fps) {
	    if (fps > gkos_config->max_fps) { /* too fast */
		target = (1.0 / gkos_config->max_fps) - (1.0 / fps);
		usleep(target * 1000000);
		fps = gkos_config->max_fps;
	    }
	}

	if (gkos_config->calc_fps)
	    printf("FPS: %f\n", fps);
    }
	
    glXSwapBuffers(dpy, window);
}

// Local variables:
// c-file-style:"cc-mode"
// End:
