/*
 * Final Program, CMPS 160, Fall 1999
 * Ben Gertzfield <ben@cse.ucsc.edu>
 *
 * This program is released under the GNU GPL, version 2 or later.
 */

/* NURBS routines */

#include <X11/Xlib.h>
#include <X11/keysym.h>
#include <gtk/gtk.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>
#include <pthread.h>
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>

#include "prog.h"

#define NUM_BANDS 16

static GLfloat scale;

static void reset_nurb(struct object *nurb);
static void nurbs_error(GLenum error_code);
static void nurbs_set_wireframe(struct object *obj, int wireframe);
static struct object *new_nurb(struct nurb_dim u, struct nurb_dim v, 
			       GLenum type, int has_weights, GLfloat *ctrl_pts);
static void free_nurb(struct object *obj);
static void set_point(struct npos *point, GLfloat x, GLfloat y, GLfloat z, GLfloat w);
static struct pos normalize_point(struct npos pt);
static struct object *do_revolution(struct nurb_curve profile);
static void sweep_sphere_update(struct object *obj, gint16 *data);

static void nurbs_error(GLenum error_code)
{
    const GLubyte *estring;
    
    estring = gluErrorString(error_code);
    fprintf(stderr, "Nurbs error: %s\n", estring);
    exit(0);
}

static void nurbs_set_wireframe(struct object *obj, int wireframe)
{
    if (wireframe) {
	obj->wireframe = 1;
	gluNurbsProperty(obj->u.nurb.nobj, GLU_DISPLAY_MODE, GLU_OUTLINE_POLYGON);
    } else {
	obj->wireframe = 0;
	gluNurbsProperty(obj->u.nurb.nobj, GLU_DISPLAY_MODE, GLU_FILL);
    }
}

/*
 * Create a new NURBS object, copying the knots and ctrl points over
 */

static struct object *new_nurb(struct nurb_dim u, struct nurb_dim v, GLenum type,
			int has_weights, GLfloat *ctrl_pts)
{
    struct object *obj;
    unsigned long num_ctrl_pts;

    obj = gkos_new_object();

    obj->type = OBJ_NURB;

    obj->u.nurb.num_rotates = 0;
    obj->u.nurb.rotates = NULL;

    obj->u.nurb.u = u;
    obj->u.nurb.v = v;

    obj->u.nurb.u.knots = (GLfloat *) malloc(sizeof(GLfloat) * u.knot_count);
    obj->u.nurb.v.knots = (GLfloat *) malloc(sizeof(GLfloat) * v.knot_count);

    memcpy(obj->u.nurb.u.knots, u.knots, sizeof(GLfloat) * u.knot_count);
    memcpy(obj->u.nurb.v.knots, v.knots, sizeof(GLfloat) * v.knot_count);

    obj->u.nurb.nobj = gluNewNurbsRenderer();

    gluNurbsProperty(obj->u.nurb.nobj, GLU_DISPLAY_MODE, GLU_FILL);
    gluNurbsCallback(obj->u.nurb.nobj, GLU_ERROR, nurbs_error);
    gluNurbsProperty(obj->u.nurb.nobj, GLU_SAMPLING_METHOD, GLU_DOMAIN_DISTANCE);
    gluNurbsProperty(obj->u.nurb.nobj, GLU_U_STEP, obj->u.nurb.u.step);
    gluNurbsProperty(obj->u.nurb.nobj, GLU_V_STEP, obj->u.nurb.v.step);

    nurbs_set_wireframe(obj, obj->wireframe);

    if (obj->u.nurb.nobj == NULL) {
	fprintf(stderr, "gluNewNurbsRenderer() returned NULL, aborting\n");
	exit(1);
    }

    obj->u.nurb.type = type;

    num_ctrl_pts = (u.knot_count - u.order) * (v.knot_count - v.order) * 
	(has_weights ? 4 : 3);

    obj->u.nurb.num_ctrl_pts = num_ctrl_pts;
    obj->u.nurb.ctrl_pts = (GLfloat *) malloc(sizeof(GLfloat) * num_ctrl_pts);
    obj->u.nurb.orig_ctrl_pts = (GLfloat *) malloc(sizeof(GLfloat) * num_ctrl_pts);

    memcpy(obj->u.nurb.ctrl_pts, ctrl_pts, sizeof(GLfloat) * num_ctrl_pts);
    memcpy(obj->u.nurb.orig_ctrl_pts, ctrl_pts, sizeof(GLfloat) * num_ctrl_pts);

    scale = 4.0 / log(256.0);

    obj->free = free_nurb;
    obj->reset = reset_nurb;

    gkos_push_object(obj);
    
    return obj;
}

static void free_nurb(struct object *obj)
{
    struct nurb nurb = obj->u.nurb;

    free(nurb.u.knots);
    free(nurb.v.knots);
    free(nurb.ctrl_pts);
    free(nurb.orig_ctrl_pts);
}


static void set_point(struct npos *point, GLfloat x, GLfloat y, GLfloat z, GLfloat w)
{
    point->x = x;
    point->y = y;
    point->z = z;
    point->w = w;
}

static struct pos normalize_point(struct npos pt)
{
    struct pos ret;
    
    ret.x = pt.x / pt.w;
    ret.y = pt.y / pt.w;
    ret.z = pt.z / pt.w;

    return ret;
}

/*
 * Revolve a NURB curve around its axis to form a NURB surface.
 *
 * Function based on code from NURBS++ library:
 * http://yukon.genie.uottawa.ca/~lavoie/software/nurbs/
 *
 */

static struct object *do_revolution(struct nurb_curve profile)
{
    int i, j, n = 9;
    struct nurb_dim u, v;
    struct object *obj;
    struct npos *point;

    GLfloat knots[12] = { 0.0, 0.0, 0.0, 0.25, 0.25, 0.50, 0.50, 0.75, 0.75,
			 1.0, 1.0, 1.0 };
    GLfloat wm, wj, r, wjm, rwjm, rwj;

    struct pos p;
    struct npos **ctrl_pts;

    ctrl_pts = (struct npos **) malloc(n * sizeof(struct npos *));
    ctrl_pts[0] = (struct npos *) malloc(n * profile.num_ctrl_pts * sizeof(struct npos));

    for(i = 1; i < n; i++)
	ctrl_pts[i] = ctrl_pts[0] + i * profile.num_ctrl_pts;

    wm = sqrt(2.0) / 2.0;

    u.order = 3;
    u.knots = knots;
    u.knot_count = 12;
    u.stride = profile.num_ctrl_pts * profile.u.stride;
    u.step = profile.u.step;

    v.order = profile.u.order;
    v.knots = profile.u.knots;
    v.stride = profile.u.stride;
    v.knot_count = profile.u.knot_count;
    v.step = profile.u.step;

    for (j = 0; j < profile.num_ctrl_pts; j++) {
	point = (struct npos *) &profile.ctrl_pts[j * v.stride];
	wj = point->w;
	p = normalize_point(*point);
	r = sqrt(p.x * p.x + p.y * p.y);
	wjm = wj * wm;
	rwjm = r * wj * wm;
	rwj = r * wj;
	p.z *= wj;

	set_point(&ctrl_pts[0][j], rwj, 0.0, p.z, wj);
	set_point(&ctrl_pts[1][j], rwjm, rwjm, p.z * wm, wjm); 
	set_point(&ctrl_pts[2][j], 0.0, rwj, p.z, wj);
	set_point(&ctrl_pts[3][j], -rwjm, rwjm, p.z * wm, wjm);
	set_point(&ctrl_pts[4][j], -rwj, 0.0, p.z, wj);
	set_point(&ctrl_pts[5][j], -rwjm, -rwjm, p.z * wm, wjm);
	set_point(&ctrl_pts[6][j], 0.0, -rwj, p.z, wj);
	set_point(&ctrl_pts[7][j], rwjm, -rwjm, p.z * wm, wjm);
	set_point(&ctrl_pts[8][j], rwj, 0.0, p.z, wj);
    }

    obj = new_nurb(u, v, GL_MAP2_VERTEX_4, 1, (GLfloat *)&ctrl_pts[0][0]);

    free(ctrl_pts[0]);
    free(ctrl_pts);

    return obj;
}

static void reset_nurb(struct object *obj)
{
    memcpy(obj->u.nurb.ctrl_pts, obj->u.nurb.orig_ctrl_pts,
	   obj->u.nurb.num_ctrl_pts * sizeof(GLfloat));

    obj->u.nurb.u.step = obj->u.nurb.v.step = gkos_config->detail_level;
    gluNurbsProperty(obj->u.nurb.nobj, GLU_U_STEP, obj->u.nurb.u.step);
    gluNurbsProperty(obj->u.nurb.nobj, GLU_V_STEP, obj->u.nurb.v.step);

    switch (gkos_config->render_style) {
    case WIREFRAME_RENDER:
	obj->texture = 0;
	nurbs_set_wireframe(obj, 1);
	break;
    case SOLID_RENDER:
	obj->texture = 0;
	nurbs_set_wireframe(obj, 0);
	break;
    case TEXMAP_RENDER:
	nurbs_set_wireframe(obj, 0);
	if (gkos_going)
	    obj->texture = gkos_new_texture(gkos_config->texmap_filename, 2);
	break;
    }

    obj->needs_update = 1;
}

/*
 * Idle update for points of sphere, based on frequency data
 *
 * based on code from example opengl_spectrum XMMS plugin
 *
 */

static void sweep_sphere_update(struct object *obj, gint16 *data)
{
    int i, j, c, y;
    GLfloat *ctrl_pts, *orig_ctrl_pts, val;
    unsigned long num_ctrl_pts;

    gint xscale[] = { 0, 1, 2, 3, 5, 7, 10, 14, 20, 28, 40, 54, 74, 101, 137, 187, 255 };

    ctrl_pts = obj->u.nurb.ctrl_pts;
    orig_ctrl_pts = obj->u.nurb.orig_ctrl_pts;
    num_ctrl_pts = obj->u.nurb.num_ctrl_pts;

    for (i = 0; i < NUM_BANDS; i++) {
	for (c = xscale[i], y = 0; c < xscale[i + 1]; c++) {
	    if (data[c] > y)
		y = data[c];
	}
	y >>= 7;

	if (y > 0)
	    val = log(y) * scale;
	else
	    val = 0;

	for (j = 0; j < 3; j++) {
	    ctrl_pts[36 + i * 8 + j] = orig_ctrl_pts[36 + i * 8 + j] +
		orig_ctrl_pts[36 + i * 8 + j] * val;
	}
    }

    obj->needs_update = 1;
}

void gkos_sweep_sphere(GLfloat r)
{
    struct nurb_curve profile;
    struct object *obj;

    GLfloat s2o2 = sqrt(2.0) / 2.0;

    GLfloat ctlpoints[5][4] = {
	{ 0.0, 0.0, r, 1.0 },
	{ -r * s2o2, 0.0, r * s2o2, s2o2 },
	{ -r, 0.0, 0.0, 1.0 },
	{ -r * s2o2, 0.0, -r * s2o2, s2o2 },
	{ 0.0, 0.0, -r, 1.0 }
    };

    GLfloat knots[8] = { 0.0, 0.0, 0.0, 0.5, 0.5, 1.0, 1.0, 1.0 };

    profile.u.knot_count = 8;
    profile.u.stride = 4;
    profile.u.order = 3;
    profile.u.knots = knots;
    profile.u.step = gkos_config->detail_level;

    profile.num_ctrl_pts = 5;
    profile.ctrl_pts = (GLfloat *) &ctlpoints[0][0];
    
    obj = do_revolution(profile);

    obj->update = sweep_sphere_update;
    obj->idle = gkos_spin_idle;
}


void gkos_draw_nurb(struct object *obj)
{
    int i;
    struct nurb nurb = obj->u.nurb;

    if (nurb.num_rotates == 0) {
	glPushMatrix();
	gluBeginSurface(nurb.nobj);
	{
	    if (obj->texture) {
		gluNurbsSurface(nurb.nobj, nurb.u.knot_count, nurb.u.knots,
				nurb.v.knot_count, nurb.v.knots, nurb.u.stride,
				nurb.v.stride, nurb.ctrl_pts, nurb.u.order,
				nurb.v.order, GL_MAP2_TEXTURE_COORD_2);
	    }

	    gluNurbsSurface(nurb.nobj, nurb.u.knot_count, nurb.u.knots,
			    nurb.v.knot_count, nurb.v.knots, nurb.u.stride,
			    nurb.v.stride, nurb.ctrl_pts, nurb.u.order,
			    nurb.v.order, GL_MAP2_NORMAL);
	    gluNurbsSurface(nurb.nobj, nurb.u.knot_count, nurb.u.knots,
			    nurb.v.knot_count, nurb.v.knots, nurb.u.stride,
			    nurb.v.stride, nurb.ctrl_pts, nurb.u.order,
			    nurb.v.order, nurb.type);
	}
	gluEndSurface(nurb.nobj);
	glPopMatrix();
    } else {
	for (i = 0; i < nurb.num_rotates; i++) {
	    glPushMatrix();
	    glRotatef(nurb.rotates[i].x, 1.0, 0.0, 0.0);
	    glRotatef(nurb.rotates[i].y, 0.0, 1.0, 0.0);
	    glRotatef(nurb.rotates[i].z, 0.0, 0.0, 1.0);
	    gluBeginSurface(nurb.nobj);
	    {
		if (obj->texture) {
		    gluNurbsSurface(nurb.nobj, nurb.u.knot_count, nurb.u.knots,
				    nurb.v.knot_count, nurb.v.knots, nurb.u.stride,
				    nurb.v.stride, nurb.ctrl_pts, nurb.u.order,
				    nurb.v.order, GL_MAP2_TEXTURE_COORD_2);
		}

		gluNurbsSurface(nurb.nobj, nurb.u.knot_count, nurb.u.knots,
				nurb.v.knot_count, nurb.v.knots, nurb.u.stride,
				nurb.v.stride, nurb.ctrl_pts, nurb.u.order,
				nurb.v.order, GL_MAP2_NORMAL);
		gluNurbsSurface(nurb.nobj, nurb.u.knot_count, nurb.u.knots,
				nurb.v.knot_count, nurb.v.knots, nurb.u.stride,
				nurb.v.stride, nurb.ctrl_pts, nurb.u.order,
				nurb.v.order, nurb.type);
	    }
	    gluEndSurface(nurb.nobj);
	    glPopMatrix();
	}
    }
}

// Local variables:
// c-file-style:"cc-mode"
// End:
