/*
 * Final Program, CMPS 160, Fall 1999
 * Ben Gertzfield <ben@cse.ucsc.edu>
 *
 * This program is released under the GNU GPL, version 2 or later.
 */

/* Object routines */

#include <X11/Xlib.h>
#include <X11/keysym.h>
#include <gtk/gtk.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>
#include <pthread.h>
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>

#include "prog.h"
#include "material.h"

struct object *gkos_first_object = NULL;
struct object *cur_object = NULL;

static void draw_model(struct object *obj);
static void draw_object(struct object *obj);

static int obj_id = 0;
static GLuint display_lists[MAX_DISPLAY_LISTS];

void gkos_init_objects(void)
{
    GLuint n, i;

    n = glGenLists(MAX_DISPLAY_LISTS);

    for (i = 0; i < MAX_DISPLAY_LISTS; i++)
	display_lists[i] = n + i;
}

void gkos_draw_all_objects(void)
{
    struct object *obj = gkos_first_object;

    if (obj == NULL) {
	fprintf(stderr, "Warning, draw_all_objects() called with no objects to draw!\n");
	return;
    }

    pthread_mutex_lock(&draw_mutex);

    while (obj != NULL) {
	draw_object(obj);
	obj = obj->next;
    } 

    pthread_mutex_unlock(&draw_mutex);
}

void gkos_reset_all_objects(void)
{
    struct object *obj = gkos_first_object;

    pthread_mutex_lock(&draw_mutex);

    while (obj != NULL) {
	if (obj->reset != NULL)
	    (*obj->reset)(obj);
	obj = obj->next;
   }

    pthread_mutex_unlock(&draw_mutex);
}

static void draw_object(struct object *obj)
{
    glPushMatrix();
    {
	if (glIsTexture(obj->texture)) {
	    glEnable(GL_TEXTURE_2D);
	    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	    glBindTexture(GL_TEXTURE_2D, obj->texture);
	}

	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, obj->material.amb_color);
	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, obj->material.dif_color);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, obj->material.spec_color);
	glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, obj->material.shininess);

	glTranslatef(obj->center.x, obj->center.y, obj->center.z);
	glRotatef(obj->rot_angle.x, 1.0, 0.0, 0.0);
	glRotatef(obj->rot_angle.y, 0.0, 1.0, 0.0);
	glRotatef(obj->rot_angle.z, 0.0, 0.0, 1.0);
	glScalef(obj->scale.x, obj->scale.y, obj->scale.z);

	if (obj->needs_update) {
	    obj->needs_update = 0;
	    obj->display_list = display_lists[0];

	    glNewList(obj->display_list, GL_COMPILE_AND_EXECUTE);
	    {
		switch (obj->type) {
		case OBJ_NURB:
		    gkos_draw_nurb(obj);
		    break;
		case OBJ_MODEL:
		    draw_model(obj);
		    break;
		}
	    }
	    glEndList();

	} else
	    glCallList(obj->display_list);

	glDisable(GL_TEXTURE_2D);
    }
    glPopMatrix();
}

void gkos_push_object(struct object *obj)
{
    struct object *temp;

    pthread_mutex_lock(&draw_mutex);

    if (gkos_first_object == NULL) {
	gkos_first_object = obj;
	pthread_mutex_unlock(&draw_mutex);
	return;
    }

    temp = gkos_first_object;

    while (temp->next != NULL)
	temp = temp->next;

    temp->next = obj;
    
    pthread_mutex_unlock(&draw_mutex);
}

void gkos_pop_object(void)
{
    struct object *first, *next;

    if (gkos_first_object == NULL) 
	return;

    pthread_mutex_lock(&draw_mutex);

    first = gkos_first_object;
    next = gkos_first_object->next;

    if (first->free != NULL)
	(*first->free)(first);

    gkos_first_object = next;

    pthread_mutex_unlock(&draw_mutex);
}

struct object *gkos_new_object(void)
{
    struct object *obj;

    obj = (struct object *) malloc(sizeof(struct object));

    if (obj == NULL) {
	fprintf(stderr, "malloc() failed in new_object(), exiting\n");
	exit(1);
    }
    obj->id = obj_id++;		/* Initialize the fields of the new object */
    obj->rot_angle.x = obj->rot_angle.y = obj->rot_angle.z = 0.0;
    obj->scale.x = obj->scale.y = obj->scale.z = 1.0;
    obj->center.x = obj->center.y = obj->center.z = 0.0;
    obj->material = gkos_get_material(DEFAULT_MATERIAL);
    obj->color.r = 0.7;
    obj->color.g = 0.7;
    obj->color.b = 0.7;

    switch (gkos_config->render_style) {
    case WIREFRAME_RENDER:
	obj->wireframe = 1;
	obj->texture = 0;
	break;
    case SOLID_RENDER:
	obj->wireframe = 0;
	obj->texture = 0;
	break;
    case TEXMAP_RENDER:
	obj->wireframe = 0;
	obj->texture = gkos_new_texture(gkos_config->texmap_filename, 2);
	break;
    }

    obj->display_list = 0;
    obj->needs_update = 1;
    obj->next = NULL;

    obj->idle = NULL;
    obj->update = NULL;
    obj->reset = NULL;
    obj->free = NULL;

    return obj;
}

void gkos_spin_idle(struct object *obj)
{
    if (obj == NULL)
	return;

    pthread_mutex_lock(&draw_mutex);

    obj->rot_angle.x = (GLfloat)((int) obj->rot_angle.x + 3 % 360);
    obj->rot_angle.y = (GLfloat)((int) obj->rot_angle.y + 2 % 360);
    obj->rot_angle.z = (GLfloat)((int) obj->rot_angle.z + 4 % 360);

    pthread_mutex_unlock(&draw_mutex);
}

static void draw_model(struct object *obj)
{

}

// Local variables:
// c-file-style:"cc-mode"
// End:
