/*
 * Final project, CMPS 160, Fall 1999
 * Ben Gertzfield <ben@cse.ucsc.edu>
 *
 * This program is released under the GNU GPL, version 2 or later.
 */

/* Function definitions and globals */

#ifndef PROG_H
#define PROG_H

#define OBJ_NURB 0
#define OBJ_MODEL 1

#define WIREFRAME_RENDER 0
#define SOLID_RENDER 1
#define TEXMAP_RENDER 2
#define ENVMAP_RENDER 3

#define INIT_WINW 400
#define INIT_WINH 400
#define INIT_NUM_LIGHTS 2
#define INIT_DETAIL 8
#define INIT_LIGHTS_MOVE 1
#define INIT_LIGHTS_CHANGE_COLOR 1
#define INIT_CALC_FPS 0
#define INIT_TDFX_MODE 0
#define INIT_RENDER_STYLE WIREFRAME_RENDER
#define INIT_MAX_FPS 99.9

#define DEFAULT_MATERIAL Gkos_Pearl

#define MAX_TEXTURES 16
#define MAX_DISPLAY_LISTS 8

#define DRAG_ROT_CAMERA 0

#define PI 3.141592653589793236482

#define NUM_FREQUENCIES 256

#define TEXMAP_FILESEL 1
#define ENVMAP_FILESEL 2

#define DEFAULT_TEXMAP "funkmotor.rgb"
#define DEFAULT_ENVMAP "env-coffee.rgb"

extern GLfloat gkos_winw, gkos_winh;
extern gchar *texture_filename;
extern pthread_mutex_t draw_mutex;

struct config {
    gint num_lights;
    gint detail_level;
    gboolean lights_move;
    gboolean lights_change_color;
    gboolean calc_fps;
    gboolean tdfx_mode;
    gint render_style;
    gfloat winw, winh;
    gfloat max_fps;
    gchar *texmap_filename;
    gchar *envmap_filename;
};

struct material {
    GLfloat amb_color[4];
    GLfloat dif_color[4];
    GLfloat spec_color[4];
    GLfloat shininess;
};

struct tex_coord {
    GLfloat s, t;
};

struct points {
    GLfloat min_x, max_x, min_y, max_y, min_z, max_z;
    unsigned long num_pts;
    struct pos *pt_array;
};

struct pos {
    GLfloat x, y, z;
};

struct npos {
    GLfloat x, y, z, w;
};

struct color {
    GLfloat r, g, b;
};

struct polys {
    GLuint num_polys;
    GLsizei *num_points;	/* num_polys wide array of points per poly */
    GLuint **points;		/* 10 x num_polys 2d array of indexes */
    struct pos *normals;	/* num_polys wide array of normal vectors */
};

struct model {
    struct pos *points;
    struct tex_coord *tex_coords;
    struct pos *pt_normals;
    struct pos max;
    struct pos min;
    struct polys *polys;
};

struct nurb_dim {
    GLint knot_count;
    GLint stride;
    GLint order;
    GLint step;
    GLfloat *knots;
};

struct nurb_curve {
    struct nurb_dim u;
    unsigned long num_ctrl_pts;
    GLfloat *ctrl_pts;
};

struct nurb {
    struct nurb_dim u, v;
    GLUnurbsObj *nobj;
    GLenum type;
    unsigned long num_ctrl_pts;
    GLfloat *ctrl_pts;
    GLfloat *orig_ctrl_pts;
    unsigned long num_rotates;
    struct pos *rotates;
};

struct object {
    unsigned long id;
    unsigned long type;
    GLuint display_list;
    int needs_update;

    struct pos scale;
    struct pos rot_angle;
    struct pos center;
    struct color color;
    struct material material;

    GLuint texture;
    int wireframe;

    union {
	struct nurb nurb;
	struct model model;
    } u;

    void (*update)(struct object *, gint16 *);
    void (*idle)(struct object *);
    void (*free)(struct object *);
    void (*reset)(struct object *);

    struct object *next;
};

struct camera {
    unsigned long id;

    GLfloat fovy;
    GLfloat near_clip, far_clip;

    struct pos rot_angle;
    GLfloat zoom;
    
    struct camera *next;
    struct camera *prev;
};

struct light {
    GLenum id;
    int num;
    GLfloat direction[4];	/* direction light comes from */
    GLfloat amb_color[4];	/* ambient color of light */
    GLfloat dif_color[4];	/* diffuse color of light */
    GLfloat spec_color[4];	/* specular (shiny) color of light */
    GLfloat color_delta[4];

    struct pos rot_angle;
    struct pos delta_angle;

    int enabled;

    void (*idle)(struct light *);
    void (*free)(struct light *);

    struct light *next;
};

/* External globals */

extern struct camera *gkos_first_camera;
extern struct camera *gkos_cur_camera;
extern struct light *gkos_first_light;
extern struct light *gkos_cur_light;
extern struct object *gkos_first_object;
extern struct object *gkos_cur_object;
extern GLuint gkos_textures[];

extern struct config *gkos_config;
extern gboolean gkos_going;

/* light.c */

struct light *gkos_new_light(void);
void gkos_enable_all_lights(void);
void gkos_enable_light(struct light *light);
void gkos_disable_light(struct light *light);
void gkos_select_light(int pickme);
void gkos_rotate_light(struct light *light, int x0, int y0, int z0, 
		  int x, int y, int z);
void gkos_pop_light(void);
void gkos_reset_all_lights(void);
void gkos_init_lights(void);

/* camera.c */

struct camera *gkos_new_camera(void);
void gkos_rotate_camera(struct camera *camera, int x0, int y0, int z0, 
			int x, int y, int z);
void gkos_zoom_camera(struct camera *camera, int z0, int z);
void gkos_setup_projection(struct camera *camera);
void gkos_pop_camera(void);

/* nurbs.c */

void gkos_sweep_sphere(GLfloat r);
void gkos_draw_nurb(struct object *obj);

/* object.c */

void gkos_init_objects(void);
void gkos_push_object(struct object *obj);
struct object *gkos_new_object(void);
void gkos_draw_all_objects(void);
void gkos_spin_idle(struct object *obj);
void gkos_reset_all_objects(void);
void gkos_pop_object(void);

/* material.c */

struct material gkos_get_material(float *mat_array);
void gkos_set_material(struct object *obj, float *mat);

/* texture.c */

GLuint gkos_new_texture(gchar *filename, int dim);
GLuint gkos_load_texture(gchar *filename);
GLuint gkos_lookup_texture(gchar *filename);
void gkos_init_textures(void);

/* about.c */

void geekosphere_about(void);

/* configure.c */

void geekosphere_configure(void);
void gkos_init_configuration(void);

#endif PROG_H

// Local variables:
// c-file-style:"cc-mode"
// End:
